"""Handles communication on the backend side between frontend and backend"""
from typing import Any, Optional, Sequence

class MissingCommand(TypeError): ...

class BackendProxy:
    backend_module: str
    backend_object: Optional[str]
    backend: Any
    def __init__(self, backend_module: str, backend_obj: Optional[str]) -> None: ...
    def __call__(self, name: str, *args: Any, **kwargs: Any) -> Any: ...
    def __str__(self) -> str: ...
    def _exit(self) -> None: ...

def run(argv: Sequence[str]) -> int: ...
def flush() -> None: ...
